from series.client.errors import SeriesClientException

from tek import logger, configurable


@configurable(client=['cli_cmd', 'cli_params'])
class HTTPCLI(object):

    def dispatch_command(self, client, cmd_name, params):
        try:
            command = getattr(client, cmd_name)
        except AttributeError:
            logger.error('No such command: {}'.format(cmd_name))
        else:
            try:
                return command(*params)
            except TypeError as e:
                logger.error(e)

    def run(self):
        cmd_name = self._cli_cmd[0]
        client = self._client
        try:
            self.dispatch_command(client, cmd_name, self._cli_params)
        except SeriesClientException as e:
            logger.error(e)
        else:
            return True

    @property
    def _client(self):
        pass

__all__ = ['HTTPCLI']



from tek.errors import TException


class MissingMetadata(TException):
    pass


class SeriesException(TException):
    pass


class InvalidDBError(SeriesException):
    def __init__(self, specifics):
        text = 'Invalid database: {}'
        super().__init__(text.format(specifics))



from tek.config import configurable
from tek import logger
from tek.tools import find

from tek_utils.sharehoster.torrent import torrent_cacher_client
from tek_utils.sharehoster.errors import ShareHosterError
from tek_utils.sharehoster.putio import PutIoClient

from series.get.handler import ReleaseHandler
from series.get.errors import SeriesDException


@configurable(torrent=['cacher'])
class TorrentHandler(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        self._pending = []
        super().__init__(releases, 5, 'torrent handler',
                                             **kw)
        self._client = torrent_cacher_client()

    def _check(self):
        super()._check()
        self._check_pending()
        self._check_error()

    def _handle(self, monitor):
        for torrent in monitor.cachable_torrents:
            msg = 'Requesting torrent download for {}…'.format(monitor.release)
            logger.info(msg)
            self._pending.append([monitor, torrent])
            try:
                torrent.request()
            except ShareHosterError as e:
                msg = 'Error requesting torrent for {}: {}'
                logger.error(msg.format(monitor.release, e))

    def _check_pending(self):
        done = []
        for monitor, torrent in self._pending:
            try:
                if torrent.cached:
                    self._releases.torrent_cached(torrent.record.id)
                    done.append([monitor, torrent])
                    msg = 'Flagging torrent as cached: {}'
                    logger.info(msg.format(monitor.release))
            except ShareHosterError as e:
                msg = 'Error checking torrent status for {}: {}'
                logger.error(msg.format(monitor.release, e))
        for item in done:
            self._pending.remove(item)

    def _check_error(self):
        errors = [t.get('id', 0) for t in self._client.transfers
                  if t.get('status') == 'ERROR']
        if errors:
            logger.info('Canceling erroneous torrents.')
            self._client.cancel_transfers(errors)
            self._client.clean_transfers()

    def _qualify(self, monitor):
        try:
            return (not monitor.downloaded and
                    monitor.has_cachable_torrents and
                    not self._requesting(monitor))
        except ShareHosterError as e:
            msg = 'Error checking torrent status for {}: {}'
            logger.error(msg.format(monitor.release, e))

    def _requesting(self, monitor):
        return find(lambda item: item[0] == monitor, self._pending)

    def _sanity_check(self):
        self._check_cacher_config()
        self._check_service_accessible()

    def _check_cacher_config(self):
        if not self._cacher:
            raise SeriesDException('No torrent cacher configured!')

    def _check_service_accessible(self):
        cacher = torrent_cacher_client()
        if not cacher.account_info:
            raise SeriesDException(
                'Couldn\'t access torrent cacher \'{}\'!'.format(self._cacher))

__all__ = ['TorrentHandler']



from tek import configurable

from tryp import _, __

from series.get.releases_facade import ReleasesFacade
from series.get.db import FileDatabase
from series.app import App
from series.get.shows_facade import ShowsFacade
from series.handler import Handler
from series.get.rest_api import RestApi


@configurable(get=['db_path', 'run', 'omit', 'auto_upgrade_db'])
class SeriesGetD(App):
    components = ['feed_poller', 'downloader', 'archiver', 'subsyncer',
                  'rest_api', 'link_handler', 'torrent_handler',
                  'library_handler', 'show_planner', 'torrent_finder',
                  'show_scheduler']

    def __init__(self):
        self._setup_db()
        super().__init__(
            'get', self._run, self._omit,
            c_args=(self.releases, self.shows), name='SeriesGetD'
        )
        self.rest_api = self._components.find(lambda a: isinstance(a, RestApi))
        self.rest_api.foreach(__.set_daemon(self))

    def _setup_db(self):
        self.db = FileDatabase(self._db_path,
                               auto_upgrade=self._auto_upgrade_db)
        self.releases = ReleasesFacade(self.db)
        self.shows = ShowsFacade(self.db)

    def load_monitors(self):
        count = self.releases.count
        self.log.info('Loaded {} releases from db.'.format(count))

    def tick(self):
        pass

    def prepare(self):
        self.load_monitors()

    def _cleanup(self):
        self.db.commit()
        self.db.disconnect()

    def activate_release(self, id):
        for c in self._components:
            if isinstance(c, Handler):
                c.activate_id(id)

__all__ = ['SeriesGetD']



import time
from threading import Thread
from datetime import datetime, timedelta

import sqlalchemy

from tek import logger
from tek.tools import find

from series.get.errors import SeriesDException


class Handler(Thread):

    def __init__(self, interval, description, cooldown=0):
        self._interval = interval
        self._description = description
        super().__init__(name=self._description)
        self._running = True
        self._initial_wait = 5
        self._failures = 0
        self._cooldown = cooldown
        self._last_check = {}

    def run(self):
        logger.info('Starting {}.'.format(self._description))
        try:
            self._sanity_check()
        except SeriesDException as e:
            logger.error('Shutting down {} with reason: {}'.format(
                self._description, e))
        else:
            self._main_loop()

    def _main_loop(self):
        next_check = datetime.now() + timedelta(seconds=self._initial_wait)
        while self._running:
            if datetime.now() > next_check:
                self._check_error_wrapper()
                next_check = self._next_check
            time.sleep(1)

    def stop(self):
        self._running = False

    def _check_error_wrapper(self):
        success = False
        try:
            success = self._check()
        except sqlalchemy.exc.InvalidRequestError as e:
            logger.debug(e)
        except Exception as e:
            self._failure(e)
        else:
            if success:
                self._success()

    def _check(self):
        with self._lock:
            current = self._current
            if current is not None:
                if hasattr(current, 'id'):
                    self._last_check[current.id] = time.time()
                self._handle(current)
                return True

    @property
    def _next_check(self):
        return datetime.now() + timedelta(seconds=self._interval)

    def _handle(self, task):
        pass

    @property
    def _current(self):
        candidates = [c for c in self._candidates if self._cool(c)]
        return find(self._qualify, candidates)

    def _cool(self, item):
        return (not hasattr(item, 'id') or time.time() -
                self._last_check.get(item.id, 0) > self._cooldown)

    def activate_id(self, id):
        if id in self._last_check:
            del self._last_check[id]

    @property
    def _candidates(self):
        raise NotImplementedError('Handler._candidates')

    def _qualify(self, candidate):
        return True

    def _commit(self):
        pass

    def _sanity_check(self):
        pass

    def _failure(self, exc):
        self._failures += 1
        import traceback
        logger.error('Error in {}:'.format(self._description))
        logger.error(exc)
        traceback.print_exc()
        if self._failures >= 20:
            msg = '20 failures in a row in {}, shutting down.'
            logger.error(msg.format(self._description))
            self.stop()

    def _success(self):
        self._failures = 0

    @property
    def _lock(self):
        raise NotImplementedError('Handler._lock')

__all__ = ['Handler']



import os
import socket

import requests

from tek import logger
from tek.config import configurable
from tek.errors import NotEnoughDiskSpace
from tek.tools import sizeof_fmt

from tek_utils.sharehoster import DownloaderFactory
from tek_utils.sharehoster.errors import ShareHosterError

from series.get.handler import ReleaseHandler


@configurable(get=['download_dir', 'min_size'])
class Downloader(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(releases, 5, 'download manager')
        self._min_failed = 0
        self._downloader_fact = DownloaderFactory()

    def _fail(self, monitor, link):
        f = monitor.failed_downloads + 1
        self._update(monitor, failed_downloads=f)
        self._releases.fail_link(link.id)

    def _handle(self, monitor):
        logger.info('Downloading {!s}.'.format(monitor.release))
        self._update(monitor, downloading=True)
        link = monitor.link
        self._download_dir.mkdir(parents=True, exist_ok=True)
        try:
            dl = self._downloader_fact(link.download_url,
                                       download_dir=str(self._download_dir))
            dl.retrieve()
        except NotEnoughDiskSpace as e:
            logger.error('Downloader: {}'.format(e))
        except (requests.RequestException, ShareHosterError,
                ConnectionError, socket.timeout) as e:
            self._fail(monitor, link)
            logger.error('Downloader: {}'.format(e))
        else:
            self._check_download(monitor, link, dl)
        finally:
            self._update(monitor, downloading=False)

    def _check_download(self, monitor, link, dl):
        ''' Verify the downloaded file's size by means of two criteria:
        The content-length header sent from the hoster must match
        exactly.
        The config parameter min_size defines what is considered to be a
        faulty file, i.e. the link checker didn't detect the file to be
        invalid, but it is small enough to just be the hoster's error
        page.
        '''
        size = dl.outfile_size
        if size < self._min_size:
            self._fail(monitor, link)
            msg = 'File smaller than required ({}): {!s}'
            logger.error(msg.format(sizeof_fmt(size), monitor.release))
        elif size != dl.file_size and dl.file_size != -1:
            self._fail(monitor, link)
            msg = 'Filesize mismatch in download of {!s}:'
            logger.error(msg.format(monitor.release))
            msg = 'Got {}, expected {}.'
            logger.error(msg.format(size, dl.file_size))
            try:
                os.unlink(dl.file_path)
            except IOError:
                msg = 'Couldn\'t delete broken download of {}.'
                logger.warn(msg.format(monitor.release))
        else:
            self._releases.update_by_id(monitor.id,
                                        download_path=str(dl.file_path))
            self._releases.mark_episode_downloaded(monitor)

    @property
    def _candidates(self):
        candidates = self._releases.downloadable
        counts = [r.failed_downloads for r in candidates]
        self._min_failed = min(counts or [0])
        return candidates

    def _qualify(self, release):
        return release.failed_downloads == self._min_failed

__all__ = ['Downloader']



from series.handler import Handler


class BaseHandler(Handler):

    def __init__(self, data, interval, description, **kw):
        self._data = data
        super().__init__(interval, description, **kw)

    @property
    def _candidates(self):
        return self._data.all

    @property
    def _lock(self):
        return self._data.lock


class ReleaseHandler(BaseHandler):

    @property
    def _releases(self):
        return self._data

    def _update(self, monitor, **data):
        self._releases.update_by_id(monitor.id, **data)


class ShowHandler(BaseHandler):

    @property
    def _shows(self):
        return self._data

__all__ = ['ReleaseHandler', 'ShowHandler']
